

/*
* Multi-threaded data steps - Master Thread Code
* <P>
* Example code for how to create a multi-threaded data step. This example is for a simple, single
* set statement input. More complex data steps may be possible, depending on the nature of the
* data step. Inputs from flat files are not supported, as there is no random access into specific
* observations.
* </P>
* <P>
* <B>WARNING:</B> You must be very careful to ensure that the data processing is accurate when
* multi-threading. Any inter-row dependencies such as use of LAG functions and retain statements
* will affect the accuracy. It is suggested that you sample and compare the output of your
* simple data step with the multi-threaded version, to ensure correct processing.
* </P>
* @author Ian J. Ghent
* @created 2009-08-23
* @modified 2010-02-09 by Ian J. Ghent
* @version 1.0
* @see multiThreadDataStep
*/

/**
* multiThreadDataStep macro
* @author Ian J. Ghent
* @param numberOfThreads Number of threads to divide the data step into
* @param inputDatasetName The input dataset which will be used for the data step
* @param partitionMethod Select HORIZONTAL, INTERLEAVED, OR INDEXED (custom needs)
* @param MinionThreadCodeLocation Location of the Minion thread code
* @param MinionThreadLogsLocation Location where the Minion thread logs are to go
* @param horizontalLeadObs In horizontal partitioning, number of obs to lead, so that
* some inter-row dependent functions such as LAG() can be
* used
* @return outputDatasetName The processes dataset with the desired name as specified
*/

%macro multiThreadDataStep(numberOfThreads, inputDatasetName, partitionMethod, MinionThreadCodeLocation, 
 MinionThreadLogsLocation, horizontalLeadObs, outputDatasetName); 

	%*Assign any needed libraries;
	libname THREAD 'C:\SAS_TempStorage\THREAD'; 

	%*Some strings used in this macro might be longer then 255 chars, so suppress the warnings; 
	%let quoteOption = %sysfunc(getoption(QUOTELENMAX));
	OPTIONS NoQuoteLenMax;

	%*pickup the sasroot path from the dictionary tables; 
	proc SQL noprint;
		SELECT path INTO :sasRoot FROM dictionary.members WHERE libname = 'SASROOT';
	QUIT;

	%*Set values for SASExecPath and SASConfigPath to reference in the systask command; 
	data _NULL_; 
		length path_name $ 256;
		path_name = "'" || trim("&sasRoot") || "\sas.exe" || "'"; 
			call symputx("SASExecPath", path_name); 
		path_name = "'" || trim("&sasRoot") || "\sasv" || scan("&sysver",1,'.') || ".cfg" || "'"; 
			call symputx("SASConfigPath", path_name); 
	run;

	%* If no number of threads is specified use the CPUCOUNT SAS system option; 
	%if &numberOfThreads. = %str() %then %do;
		%put NOTE: **** THREADS NOT SET AS OPTION, RETREIVING FROM SYSTEM CPUCOUNT OPTION ****; 
		%let numberOfThreads = %sysfunc(getoption(CPUCOUNT));
	%end; 

	%* output selected options to the log ;
	%put NOTE: **** MULTI-THREADED DATA STEP INFORMATION ****;
	%put NOTE- **** NUMBER OF THREADS: &numberOfThreads. ****;
	%put NOTE- **** DATASET TO PROCESS: &inputDatasetName. ****;
	%put NOTE- **** DATASET TO OUTPUT: &outputDatasetName. ****;
	%put NOTE- **** PARTITION METHOD: &partitionMethod. ****;
	%put NOTE- **** MINION THREAD CODE: &MinionThreadCodeLocation. ****;
	%if &partitionMethod. = HORIZONTAL and &horizontalLeadObs. ^= %str() %then
		%put NOTE- **** HORIZONTAL LEAD OBSERATIONS: &horizontalLeadObs. ****;;
	%put; 

	%*Clean out any existing thread datasets so we dont get confused;
	proc datasets library = THREAD nolist nodetails nowarn;
		delete thread_part: thread_not_started: thread_running: thread_completed:; 
	quit; 

	%*Variable initialization; 
	%let threadList = ;
	%let firstRecordPointer = 1;
	%let firstObs = .;
	%let lastObs = .;

	%*Loop through each thread, writing out information datasets and executing the thread;
	%do i = 1 %to &numberOfThreads.; 

		%*If horizontal partitioning is selected, calculate the first/last obs for each thread to process;
		%if (&partitionMethod. = HORIZONTAL) %then %do;
			%*There are lots of quick ways to pick up the number of observations from the data descriptor.; 
			%*However, only a real count is accurate in all situations. Proc sql is used to provide multi-threaded capabilities;
			proc SQL noprint threads;
				SELECT count(*) INTO :recordCount
					FROM &inputDatasetName.;
			QUIT; 

			%let recordsPerThread = %sysevalf(&recordCount./&numberOfThreads., floor);
			%let firstObs = &firstRecordPointer.;
			%if &i. > 1 and &horizontalLeadObs. ^= %str() %then %let firstObs = %eval(&firstObs - &horizontalLeadObs.);
			%let lastObs = %eval(&firstRecordPointer. + &recordsPerThread. - 1); 

			%*The last thread picks up any additional records that did not divide evenly; 
			%if &i. = &numberOfThreads. %then %let lastObs = &recordCount.;
			%let firstRecordPointer = %eval(&firstRecordPointer. + &recordsPerThread);
		%end; 

		data THREAD.thread_not_started&i.;
			thread_id = &i.;
			number_of_threads = &numberOfThreads.;
			dataset_name = "&inputDatasetName.";
			partition_method = "&partitionMethod.";
			first_obs = &firstObs.;
			last_obs = &lastObs.;
			where_clause = '';
			start_datetime = .;
			finish_datetime = .;
			run_time_minutes = .;
				format start_datetime finish_datetime datetime.; 
		run; 

		%*Build up the sas command (easier using a data step);
		data _NULL_;
			sascmd = "&SASExecPath." || " -sysin '" || "&MinionThreadCodeLocation." || "'" || 
			" -config " || "&SASConfigPath." || " -log '&MinionThreadLogsLocation.\thread&i..log" || "'" || 
			" -noterminal -nostatuswin -sysparm &i."; 
			call symput('sascmd', sascmd); 
		run; 

		%*Execute the command;
		SYSTASK command "&sascmd." taskname = thread&i. status = threadRC&i.; 

		%*Build up a list of all the threads by taskname;
		%let threadList = &threadList. thread&i.; 
	%end;

	%put NOTE: **** WAITING FOR THREADS TO FINISH ****; 

	%*Tell SAS to wait for all threads to finish before continuing;
	WAITFOR _all_ &threadList.; 
	%put NOTE: **** THREADS ARE FINISHED ****;

	%*Error check and recombine all datasets together;
	%do i = 1 %to &numberOfThreads.; 
		%*Check to make sure every thread completed successfully; 
		%if (&&threadRC&i.. ^= 0) %then %do;
			%put ERROR: **** THREAD PROCESS &I. FAILED, PLEASE CHECK THE THREAD LOG ****; 
			%ABORT; 
		%end; 
		%if (&i. = 1) %then %do;
			data &outputDatasetName.;
			set THREAD.thread_part1;
			run;
		%end;
			%else %do;
				proc append base = &outputDatasetName. data = THREAD.thread_part&i.
				%if (&horizontalLeadObs. ^= %str()) %then (firstobs = %eval(&horizontalLeadObs. + 1));;; 
				run; 
			%end; 
	%end; 

	%put NOTE: **** REJOIN COMPLETE ****;

	%*Clean up the partitioned datasets (but not the informational datasets);
	proc datasets library = THREAD nolist nodetails nowarn; 
		delete thread_part:; 
	quit; 

	%*Reset changed options;
	OPTIONS &quoteOption.; 

%mend multiThreadDataStep; 

